import { utils } from "zebulon-controls";
import { getFunction } from "./compute.meta"; // ----------------------------
//  filters
//  ---------------------------

export var filterFunction = function filterFunction(column, params, data, updatedRows) {
  var facc = function facc(row) {
    var f = column.foreignKeyAccessorFunction || column.accessorFunction;

    if (column.id === "status_") {
      var status = (updatedRows || {})[row.index_] || {};
      return 1 * (status.updated_ || 0) + 2 * (status.new_ || 0) + 4 * (status.deleted_ || 0) + 8 * ((status.errors || {
        n_: 0
      }).n_ > 0);
    }

    if (!f) {
      if (column.dataType === "date" || column.dataType === "month" || column.dataType === "year") {
        f = function f(_ref) {
          var row = _ref.row,
              column = _ref.column;
          return new Date(row[column.id]);
        };
      } else {
        f = function f(_ref2) {
          var row = _ref2.row,
              column = _ref2.column;
          return row[column.id];
        };
      }
    }

    var v = f({
      row: row,
      column: column,
      params: params,
      status: (updatedRows || {})[row.index_],
      data: data
    });

    if (column.select && v !== null && typeof v === "object" && !utils.isDate(v)) {
      return v.value;
    }

    return v;
  };

  if (column.dataType === "date" || column.dataType === "month" || column.dataType === "year") {
    if (!utils.isNullOrUndefined(column.v)) {
      column.v.value = new Date(column.v.value);
    }

    if (!utils.isNullOrUndefined(column.vTo)) {
      column.vTo.value = new Date(column.vTo.value);
    }
  }

  if (column.id === "status_") {
    column.f = function (row) {
      var vRow = facc(row);
      return vRow && (vRow & column.v) !== 0;
    };
  } else if (column.filterType === "values") {
    column.f = function (row) {
      return column.v[facc(row)] !== undefined;
    };
  } else if (column.dataType === "boolean") {
    column.f = function (row) {
      return (facc(row) || false) === column.v.value;
    };
  } else if (column.filterType === "=") {
    column.f = function (row) {
      return facc(row) === column.v.value;
    };
  } else if (column.filterType === ">=") {
    column.f = function (row) {
      return facc(row) >= column.v.value;
    };
  } else if (column.filterType === "between") {
    column.f = function (row) {
      return (!utils.isNullOrUndefined(column.vTo) ? facc(row) <= column.vTo.value : true) && (!utils.isNullOrUndefined(column.v) ? facc(row) >= column.v.value : true);
    };
  } else if (column.filterType === "<=") {
    column.f = function (row) {
      return facc(row) <= column.v.value;
    };
  } else if (column.filterType === "starts (case sensitive)") {
    column.f = function (row) {
      return String(facc(row) || "").startsWith(String(column.v.value || ""));
    };
  } else if (column.filterType === "contains") {
    column.f = function (row) {
      return String(facc(row) || "").indexOf(String(column.v.value || "")) !== -1;
    };
  } else if (column.filterType === "starts") {
    var v = String(column.v.value).toUpperCase() || "";

    column.f = function (row) {
      return String(facc(row) || "").toUpperCase().startsWith(v);
    };
  }
}; // -----------------------------

export var filtersFunction = function filtersFunction(filters, params, data, updatedRows) {
  if (!filters) {
    return function (x) {
      return x;
    };
  }

  var f = Object.values(filters).concat((params || {}).filter_ || []).filter(function (filter) {
    if (filter.dataType === "boolean" && filter.v.value === null) {
      return false;
    }

    return !utils.isNullOrUndefined(filter.v) || filter.filterType === "between" && !utils.isNullOrUndefined(filter.vTo);
  }).map(function (filter) {
    filterFunction(filter, params, data, updatedRows);
    return filter;
  });

  if (!f.length) {
    return function (row) {
      return !((updatedRows || {})[row.index_] || {}).hidden_;
    };
  }

  return function (row) {
    return !((updatedRows || {})[row.index_] || {}).hidden_ && f.reduce(function (acc, filter) {
      return acc && filter.f(row);
    }, true);
  };
}; // -----------------------------

export var getFilters = function getFilters(columns, filters) {
  if (!columns.length) {
    return filters;
  }

  return columns.filter(function (column) {
    return column.v;
  }).reduce(function (acc, column) {
    var v = column.v.value,
        vTo = column.vTo.value,
        filterType = column.filterType;

    if (filters[column.id]) {
      v = filters[column.id].v.value;
      vTo = filters[column.id].vTo.value;
      filterType = filters[column.id].filterType;
    }

    if (!utils.isNullOrUndefined(v) || column.filterType === "between" && !utils.isNullOrUndefined(column.vTo.value)) {
      acc[column.id] = {
        id: column.id,
        dataType: column.dataType,
        filterType: filterType,
        v: v,
        vTo: vTo,
        accessor: column.accessor,
        accessorFunction: column.accessorFunction
      };
    }

    return acc;
  }, {});
}; // ----------------------------
//  sorts
//  ---------------------------

export var sortsFunction = function sortsFunction(sorts) {
  return function (rowA, rowB) {
    return sorts.reduce(function (acc, sort) {
      var accessor = sort.accessor || function (_ref3) {
        var row = _ref3.row;
        return row[sort.id];
      };

      if (acc === 0) {
        var a = accessor({
          row: rowA
        }),
            b = accessor({
          row: rowB
        });
        acc = ((a > b) - (b > a)) * (sort.direction === "asc" ? 1 : -1);
      }

      return acc;
    }, 0);
  };
}; // -----------------------------------------

export var getSorts = function getSorts(columns) {
  var sorts = columns.filter(function (column) {
    return column.sort !== undefined;
  }).map(function (column) {
    var fNullValue = function fNullValue(x) {
      return utils.nullValue(x, "");
    };

    if (column.dataType === "number" || column.dataType === "boolean") {
      fNullValue = function fNullValue(x) {
        return utils.nullValue(x, -Infinity);
      };
    } else if (column.dataType === "date") {
      fNullValue = function fNullValue(x) {
        return utils.nullValue(x, new Date(null));
      };
    } // else if (column.dataType === "string") {
    //   nullValue = "";
    // }


    return {
      id: column.id,
      sortOrder: column.sortOrder,
      direction: column.sort,
      accessor: column.sortAccessorFunction || column.accessorFunction ? function (message) {
        return fNullValue(column.accessorFunction(message));
      } : function (_ref4) {
        var row = _ref4.row;
        return fNullValue(row[column.id]);
      }
    };
  });
  sorts.sort(function (a, b) {
    return (a.sortOrder > b.sortOrder) - (b.sortOrder > a.sortOrder);
  });
  return sorts;
};